from utils import fslstats, niigz, getnibdata, rats as ratnames, pats as patnames,\
                    get_pat_contname, get_pool_letter, get_pat_maskpath,\
                    fslmaths, imrm, pat_gmwm_masks,\
                    vols, get_voldirs, nvols, parse_voldir, get_voxs
                  
import numpy as np
from matplotlib import pyplot as plt
import os.path as op
from itertools import product

import scipy.stats
from scipy.special import comb
from statsmodels.stats.multicomp import pairwise_tukeyhsd
from scipy.stats import pearsonr
import pandas
from statsmodels.stats.anova import AnovaRM
from sys import exit

rois = ['ic', 'ig', 'ol']

def premake_pat_roiXgmwm_masks(delete=False, forceremake=False):
    for pat in patnames:
        roipaths = [get_maskpath(roi, pat) for roi in rois + ['co']]
        p2masks = op.dirname(roipaths[0])
        for gmwm in pat_gmwm_masks:
            p2gmwm = op.join(p2masks, gmwm)
            roisXgmwm = [f'{roipath}-{gmwm}-automk' for roipath in roipaths]
            
            for roipath, roiXgmwm in zip(roipaths, roisXgmwm):
                if not op.exists(roipath+niigz):
                    continue
                if forceremake or not op.exists(roiXgmwm+niigz):
                    fslmaths(roipath, f'-mas {p2gmwm}', roiXgmwm)
                if delete:
                    imrm(roiXgmwm)
     
def get_maskpath(roi, subname, run=None):
    if run is None:
        maskpath = get_pat_maskpath(subname, roi)
    else:
        if roi == 'ic':
            mname = 'infarctcoremask'
        elif roi == 'ig':
            mname = 'infarctgrowthmask'
        elif roi == 'ol':
            mname = 'oligemiamask'
        elif roi == 'co':
            mname = 'mask_contralateral'
        else:
            mname = roi
        maskpath = f'mcao/{subname}/{mname}r{run}'
    return maskpath

def mapname_from_poolname(modnum, poolname, run=None):    
    mapname = None
    if run is None:
        if poolname == 'convmt':
            mapname = 'ssconv_1'
        if poolname == 'amide':
            mapname = f'aptstar{modnum}_1'
        if modnum==3 and poolname == 'noemt':
            mapname = 'noestar3_1'
        elif modnum==4 and poolname == 'noe':
            mapname = 'noestar4_1'
        elif modnum==4 and poolname == 'mt':
            mapname = 'ssstar_1'
    else:
        poollabel = get_pool_letter(modnum, poolname)
        mapname = f'CESTRstar_{poollabel}'
    return mapname
        
def get_cestrpath(modnum, poolname, subname, run=None):
    mapname = mapname_from_poolname(modnum, poolname, run)
    
    if run is None:
        cpath_mine = f'pat/__fit_{subname}_{modnum}pool/CESTRstar_{get_pool_letter(modnum,poolname)}'
        cpath_mine_copy = f'pat/cest_maps_copy-__fit_/{subname}/{mapname}'
        cpath = cpath_mine_copy
    else:
        cpath = f'mcao/__fit_{subname}_run{run}_{modnum}pool/{mapname}'
    return cpath

def get_cestr_voxs(modnum, poolname, roi, subname, run=None, gmwm=None):
    cestr = get_cestrpath(modnum, poolname, subname, run)
    mask = get_maskpath(roi, subname, run)
    mask = mask if not gmwm else f'{mask}-{gmwm}-automk'
    vals = get_voxs(cestr, mask)
    return vals[~np.isnan(vals)]

def meanarr_from_voxslist(voxslist, nanfilt=True):
    meanarr = np.array( [np.nanmean(voxs) for voxs in voxslist] )
    if nanfilt:
        return meanarr[~np.isnan(meanarr)]
    else:
        return meanarr

def rmeanarr_from_voxslist(voxslist, comeans):
    rvoxs = [voxs/comean for voxs, comean in zip(voxslist, comeans)]
    return meanarr_from_voxslist(rvoxs)
    
#%%stats
def multcomp_bonferroni(subjvals, doBonferroni=True):
    nsets = len(subjvals)
    numcomparisons = int(comb(nsets, 2))
    FWER = 0.05
    pcrit = FWER / numcomparisons if doBonferroni else FWER
    
    tbl = np.zeros([numcomparisons, 8])
    crow = 0
    
    for idx1, valsg1 in enumerate(subjvals):
        for idx2, valsg2 in enumerate(subjvals):
            if idx1 >= idx2:
                continue

            _, pval = scipy.stats.ttest_ind(valsg1, valsg2,
                                            equal_var=False,
                                            nan_policy='omit',
                                            alternative='two-sided')
            m1,ci1,_,_ = getstats(valsg1)
            m2,ci2,_,_ = getstats(valsg2)
            tbl[crow, :] = np.array([idx1, idx2,
                                     m1, float(ci1),
                                     m2, float(ci2),
                                     pval<pcrit, pval], dtype=object)
            crow += 1
    return tbl

def mixed_effects(all_voxs, M0=0, usevars=None, mixed=True):    
    Yi = np.array([voxs.mean() for voxs in all_voxs])
    if not usevars:
        Vi = np.array([voxs.var(ddof=1)/len(voxs) for voxs in all_voxs])
    else:
        Vi = np.array([usevar/len(voxs) for usevar, voxs in zip(usevars, all_voxs)])
    Wi = 1/Vi
    M = (Wi*Yi).sum() / Wi.sum()
    df = len(Yi) - 1
    
    #Estimate T2 and recalc if mixed effects model
    if mixed:
        Q = ( (Yi - M)**2/Vi ).sum()
        C = Wi.sum() - (Wi**2).sum()/Wi.sum()
        T2 = (Q - df) / C
        
        Wi = 1/(Vi + T2)
        M = (Wi*Yi).sum() / Wi.sum()
    
    V_M = 1/Wi.sum()
    SE_M = np.sqrt(V_M)
    #Z = abs(M - M0)/SE_M
    #pp = stats.norm.sf(Z) * 2
    
    t = abs(M - M0)/SE_M
    pt = (1 - scipy.stats.t.cdf(abs(t), df)) * 2    
    return pt

#analyses based on provided values across subjects
def getstats(setvals):
    num = setvals.size
    sd = setvals.std(ddof=1)
    sem = sd/num**0.5
    ci = scipy.stats.t.ppf(0.975, num-1) * sem

    return np.mean(setvals), ci, num, sd**2

def get_var_ofratio(voxsroi, voxsco):
    def rmidxsfromboth(l1, l2, idxs):
        idxs.sort(reverse=True)
        [l1.pop(idx) for idx in idxs]
        [l2.pop(idx) for idx in idxs]
        return l1, l2
    
    vroi = voxsroi.copy()
    vco = voxsco.copy()
    
    idx2rm = [idx for idx in range(len(vroi)) if np.any(np.isnan(vroi[idx]))]
    vroi, vco = rmidxsfromboth(vroi, vco, idx2rm)

    idx2rm = [idx for idx in range(len(vroi)) if len(vroi[idx])<3]
    vroi, vco = rmidxsfromboth(vroi, vco, idx2rm)

    vroi = [vs[~np.isnan(vs)] for vs in vroi]
    vco = [vs[~np.isnan(vs)] for vs in vco]
    rvars = [roi.mean()**2/co.mean()**2 * (roi.var(ddof=1)/roi.mean()**2 + co.var(ddof=1)/co.mean()**2) for roi,co in zip(vroi, vco)]
    rvoxs = [roi/co.mean() for roi, co in zip(vroi, vco)]
    
    idx2rm = [idx for idx in range(len(rvars)) if np.any(np.isnan(rvars[idx])) or np.any(np.isnan(rvoxs[idx])) or np.nanmean(rvoxs[idx])<1e-3]
    rvars, rvoxs = rmidxsfromboth(rvars, rvoxs, idx2rm)
        
    return rvars, rvoxs

def clip4mt(l1,l2,l3):
    return l1[4:], l2[4:], l3[4:]
#%%
#Results are for the animal data.

runs = [1,2]
gmwms = [None]
subnames = ratnames

view_figure = 6 # Set this to 4 to load Fig 4 data, or 6 for Fig 6 data.

if view_figure == 4:
    #---------------------------------FIGURE 4--------------------------------#
    #noe and mt
    modnums_and_poolnames = [(3,'noemt'), (4,'noe'), (4,'mt')]
    subplotdims = [2,3]
    #-------------------------------------------------------------------------#
elif view_figure == 6:
    #------------------------------FIGURE 6 (animal data)---------------------#
    #apt and noe
    modnums_and_poolnames = [(4,'amide'), (4,'noe')]
    subplotdims = [2,2]
    #-------------------------------------------------------------------------#
    
plt.close('all')
xlab = ['ischaemic\ncore','infarct\ngrowth','oligaemia']

f,axarr = plt.subplots(*subplotdims)
f.set_size_inches([17,7.5])
axarr = axarr.flatten()
axidx = 0
globalmin, globalmax = 5,-5

valsdict = {}

runs_and_gmwms = list(product(runs, gmwms))
allcovs = []
for run, gmwm in runs_and_gmwms: 
    for modnum, poolname in modnums_and_poolnames:
        if not op.exists(get_cestrpath(modnum, poolname, subnames[0], run)+niigz):
            continue
        
        print('\n', modnum,'pools', poolname, f'r{run}', gmwm)
        
        spec = {'run': run,
                'gmwm': gmwm,
                'poolname': poolname,
                'modnum': modnum,  
                }
                
        #
        #Read vals. compute manually here as it's quicker than re-reading images
        #
        voxs_co = [get_cestr_voxs(**spec, roi='co', subname=sub) for sub in subnames]
        voxs_ic = [get_cestr_voxs(**spec, roi='ic', subname=sub) for sub in subnames]
        voxs_ig = [get_cestr_voxs(**spec, roi='ig', subname=sub) for sub in subnames]
        voxs_ol = [get_cestr_voxs(**spec, roi='ol', subname=sub) for sub in subnames]
        if poolname == 'convmt':
            voxs_ic,voxs_ig,voxs_ol=clip4mt(voxs_ic,voxs_ig,voxs_ol)
            voxs_co,_,_=clip4mt(voxs_co,voxs_co,voxs_co)
        
        means_co = meanarr_from_voxslist(voxs_co)
        means_ic = meanarr_from_voxslist(voxs_ic)
        means_ig = meanarr_from_voxslist(voxs_ig)
        means_ol = meanarr_from_voxslist(voxs_ol)
        
        unfilt_means_ic = meanarr_from_voxslist(voxs_ic, nanfilt=False)
        unfilt_means_co = meanarr_from_voxslist(voxs_co, nanfilt=False)
        
        rmeans_ic = rmeanarr_from_voxslist(voxs_ic, means_co)
        rmeans_ig = rmeanarr_from_voxslist(voxs_ig, means_co)
        rmeans_ol = rmeanarr_from_voxslist(voxs_ol, means_co)

        #
        #ANOVA using rmeans
        #
        _, anovap = scipy.stats.f_oneway(rmeans_ic, rmeans_ig, rmeans_ol)
       
        #print anova result
        if anovap < 0.05:
            print('\tANOVA: * p=', round(anovap, 4))
        else:
            print('\tANOVA: ns (p=',np.round(anovap,4),')')
            
        #
        #Post-hoc testing if applicable
        #
        if anovap < 0.05:
            tukey = False
            if tukey:
                df = pandas.DataFrame({'rcestr':np.concatenate([rmeans_ic,rmeans_ig,rmeans_ol]),
                                       'roi': np.repeat(['ic','ig','ol'], [rmeans_ic.size,rmeans_ig.size,rmeans_ol.size])
                                       })
                r = pairwise_tukeyhsd(endog=df['rcestr'],groups=df['roi'])
                if np.any(r.reject):
                    print(r)
            else:
                mctbl = multcomp_bonferroni([rmeans_ic, rmeans_ig, rmeans_ol],False)
                print('\t\tPost-hoc:')
                for r in mctbl:
                    if not r[-2]: continue
                    print('\t\t  *',[np.round(n, 4) for n in r]) 
        
        #
        #Print means and CIs, for reporting and plotting
        #
        m_ic, ci_ic, _, _ = getstats(rmeans_ic)
        m_ig, ci_ig, _, _ = getstats(rmeans_ig)
        m_ol, ci_ol, _, _ = getstats(rmeans_ol)
        
        valskey =f'{run}-{gmwm}-{modnum}-{poolname}' 
        valsdict[valskey + '-means'] = np.array([m_ic, m_ig, m_ol])
        valsdict[valskey + '-cis'] = np.array([ci_ic, ci_ig, ci_ol])
        
        #print
        print('\tROI mean and 95% CI:')
        print('','', 'ic=', m_ic, '± '+ str(ci_ic), sep='\t')
        print('','', 'ig=', m_ig, '± '+ str(ci_ig), sep='\t')
        print('','', 'ol=', m_ol, '± '+ str(ci_ol), sep='\t')
            
        #
        #Mixed effects using rvoxs (it computes weights incl var)
        #
        #    V2    #     
        #calc and pop nan elements beforehand
        vars_ic, rvoxs_ic = get_var_ofratio(voxs_ic, voxs_co)
        vars_ig, rvoxs_ig = get_var_ofratio(voxs_ig, voxs_co)
        vars_ol, rvoxs_ol = get_var_ofratio(voxs_ol, voxs_co)
        #
        pmx_ic = mixed_effects(rvoxs_ic, usevars = vars_ic, M0=1)
        pmx_ig = mixed_effects(rvoxs_ig, usevars = vars_ig, M0=1)
        pmx_ol = mixed_effects(rvoxs_ol, usevars = vars_ol, M0=1)
        #print mixed effects results for ic/ig/ol
        print('\tMixed effects wrt contralateral:')
        for i, pval in enumerate([pmx_ic, pmx_ig, pmx_ol]):
            marker = '*' if pval < 0.05 else ''
            print('\t\t', f'{marker} {rois[i]}:', round(pval,4)) 
                
        #
        #report contralateral COV [pats, rats run 1]
        #
        if run == 1 or run is None:
            co_covs = [np.nanstd(voxs, ddof=1)/np.nanmean(voxs)*100 for voxs in voxs_co]
            meancov = round(np.nanmean(co_covs),1)
            allcovs.append(meancov)
        
        #
        #CNR
        #
        cnrs = (unfilt_means_ic - unfilt_means_co)/(unfilt_means_ic + unfilt_means_co)
        print('\tCNR',np.nanmean(cnrs)*100,'% +-', np.nanstd(cnrs,ddof=1)*100,'%')
        
        #
        #Plotting
        #
        ymin = np.min([m_ic-ci_ic, m_ig-ci_ig, m_ol-ci_ol])
        if ymin < globalmin and poolname!='convmt': globalmin = ymin
        ymax = np.max([m_ic+ci_ic, m_ig+ci_ig, m_ol+ci_ol])
        if ymax > globalmax and poolname!='convmt': globalmax = ymax
        
        ax = axarr[axidx]
        ax.bar(xlab,
               [m_ic, m_ig, m_ol],yerr=[ci_ic, ci_ig, ci_ol],
               color='darkgrey', capsize=3)
        if 1:
            ax.set_title(f'r{run} {modnum} {poolname} {gmwm} cov{meancov}' + \
                         f'\nic{pmx_ic<0.05 }/ig{pmx_ig<0.05}/ol{pmx_ol<0.05}'*0)
        ax.set_xlim(-0.5,2.5)
        ax.plot([-10,10], [1,1], 'k:',lw=1)
        ax.set_xticklabels(xlab,rotation=45)
        
        if anovap < 0.05:
            for r in range(mctbl.shape[0]):
                row = mctbl[r,:]
                if not row[-2]: continue
                ax.plot(np.sort(row[0:2]), [1.05 + r/35]*2, 'k-|', lw=1)
                
        #add stars
        for i, (pmx,mean) in enumerate(zip([pmx_ic, pmx_ig, pmx_ol],[m_ic, m_ig, m_ol])):
            if not pmx < 0.05:
                continue
            ax.plot([i+0.3], [mean+0.015], 'ko', ms=4)
            
        if poolname == 'noemt':
            ylab = 'relative 3-pool NOER*'
        elif poolname == 'mt':
            ylab = 'relative pMTR*'
        elif poolname == 'noe':
            ylab = 'relative 4-pool NOER*'
        elif poolname == 'convmt':
            ylab = 'relative MTR'
        elif poolname == 'amide':
            ylab = 'relative APTR*'
        else:
            ylab = f'relative {poolname} CESTR*'
        ax.set_ylabel(ylab)
        
        axidx += 1

print('Overall COV=', round(np.nanmean(allcovs)),round(np.nanstd(allcovs,ddof=1)))

for ax in axarr:
    pass
    ax.set_ylim(globalmin*0.99, globalmax*1.01)
#f.tight_layout()

#
#Side by side for NOE & APT
#
if modnums_and_poolnames == [(4,'amide'), (4,'noe')]:
    f,axs = plt.subplots(1,2)
    
    for idx, run in enumerate(runs):
        key = f'{run}-{"None"}-4-amide'
        apt_means = valsdict[key + '-means']
        apt_cis  = valsdict[key + '-cis']
        
        key = f'{run}-{"None"}-4-noe'
        noe_means = valsdict[key + '-means']
        noe_cis  = valsdict[key + '-cis']
        
        ax = axs[idx]
        ax.bar([1,3.5,6], apt_means, 1, yerr=apt_cis, color='lightsteelblue', label='APTR*')
        ax.bar([2,4.5,7], noe_means, 1, yerr=noe_cis, color='grey', label='NOER*')
        
        ax.plot([-10,10], [1,1], 'k:',lw=1)
        ax.set_xlim(0, 8)
        ax.set_ylim(0.5, 1.5)
        ax.set_xticks([1.5, 4 ,6.5])
        ax.set_xticklabels(xlab,rotation=45)
        ax.set_ylabel('relative CESTR*')
        ax.legend()